comment ^
--------------------- printx macros -------------------------
   copyright (C) 2008 Jim Giordano, free for all uses public, private, and commercial.
      copies multiple strings to a buffer, optionally converting dword, word, or byte values
      to text using keywords.  Copy byte directly if immediate value < 256 e.g.  ,13,10

   The form is:   printx buffer-address,string1,string2,"literal string",....
   
   printx expects the parameters to be actual string names, string literals,
       or registers pointing at strings.

    e.g.
         .data
         str1 "aaa",0
         str2 "bbb",0
         str3 "ccc",0
         mybuff db 100 dup (?)
         .code
         printx addr mybuff,"My Strings - ",str1,",",str2,",",str3

         will concatenate the three strings to the buffer mybuff with a comma between.
             mybuff = "My Strings - aaa,bbb,ccc",0

    There also exists the macro printxa, which appends to previous buffer used by printx.
    
         The location of the final zero byte terminator is kept in the dword variable named
         printxptr, so that additional strings may be concatenated without reference to the buffer
         
         e.g.  printxa str4       ; will copy str4 after str3 in mybuff.            
    
    Strings may be copied using the address in a register, e.g.
    
          mov eax,offset str2
          printxa eax         ; will append str2 to the current buffer.
          
    There are special meaning characters-
    
        /  output crlf (13,10).  Multiple slashes acceptable eg printxa ////,"test"      

    Occasionally, you will want to print a string pointed to by a variable. To do
      so, use the actual address keyword "aa", e.g.
    
        m2m myptr,offset str2
        printxa "printing str2 using actual address=",aa myptr  

    Additional keywords for the following conversions available:

        dd signed decimal
        du unsigned decimal
        hx lower case hex
        Hx upper case hex
        aa (value represents an actual address of a string)
        xx spaces  

        e.g.
        
        printxa "Test, eax=",dd eax," hex=",hx eax
        printxa ///      ; append 3 carriage return/linefeeds (13,10)
        printxa "hx bl = ",hx bl,/  ; print hex of a byte in bl
        printxa "du bx = ",du bx,/  ; print unsigned decimal of a word in bx
        printxa "Hx ebx = ",Hx ebx  ; print hex of a dword in ebx in upper case
        printxa dd myvar   ; prints the signed decimal value of the variable myvar.
                           ;  the size is determined by its definition.
        printxa xx 4       ; prints 4 spaces                   
        
    In addition to the printxptr variable, containing the address of the zbt,
        the variable printxsize contains the number of characters in the buffer.
                                   

    There is an additional macro available, msgx, that displays the current string
    in a message box, e.g.   msgx addr mybuff,"This is a test."  Or if the buffer
    already has strings in it, just   msgx   with no parameters will display it.        

end comment ^
.data
printxname  db "printx",0
align 4
printxstart dd 0    ; start of output buffer as defined by printx or printxf
printxptr   dd 0    ; next available spot for appending more text
printxsize  dd 0    ; current number of characters in the buffer
printxeflg  db 0
printxerr1  db "Error in printx: invalid input address",0
printxerr2  db "printx called without setting start address",0
printxerr3  db "Error in printx: invalid output address",0
.code
align 16
printxproc: ; proc ; the main printx procedure called by the printx macro
    ; all registers are saved by printxa macro
    ; edi = output address
    ; dh  = flag, 0=string, address in esi
    ;             1=signed dword in eax, convert to ascii
    ;             2=byte (no conversion, value in al)
    ;             3=unsigned dword in eax, convert to ascii
    ;             4=convert binary in eax to lower case hex ascii string (dl=bytes to convert)
    ;             5=convert binary in eax to upper case hex ascii string (dl=bytes to convert)
    ;             6=new line. (number of newlines in dl)
    ;             7=floating point  (not done yet)
    ;             8=spaces (number of spaces in eax)
    ;            
    mov edi,printxptr    ;; where we will store the results
    .if edi ; has the user set an address?
        mov printxeflg,0 ;Install per-thread error handler from http://www.winasm.net/forum/index.php?showtopic=2082
        assume fs:nothing ;MASM assumes the use of this register to be ERROR by default
        push offset PrintxExceptionHandler
        push fs:[0]
        mov fs:[0],esp
        .if dh==0    ; a zbt string with address in esi
            .repeat
                movzx eax,byte ptr [esi]
                add esi,1
                mov [edi],al
                add edi,1
            .until al==0
            dec edi
        .elseif dh==1 || dh==3 ; flag to convert dword to text
            ; adapted from masm32 lib's dwtoa but can convert to both signed and unsigned values
            or eax,eax  ; test for zero and negative values
            .if Zero?
                mov word ptr [edi],30h  ; "0" and zero byte terminator
                inc edi     ; point at terminator
            .else
                .if Sign?    
                    .if dh==1    ; check signed
                        mov byte ptr [edi],'-'
                        neg eax
                        add edi, 1
                    .endif
                .endif
                mov ecx, 0CCCCCCCDh ; 8 * 1/10
                mov esi, edi    ; save start of string
                .repeat
                    mov ebx,eax
                    mul ecx
                    shr edx, 3
                    mov eax,edx
                    lea edx,[edx*4+edx]
                    add edx,edx
                    sub ebx,edx
                    add bl,'0'
                    mov [edi],bl
                    add edi, 1
                .until eax==0
                mov byte ptr [edi], 0   ; terminate the string
                mov ecx,edi
                .repeat                 ; We now have all the digits, but in reverse order.
                  sub ecx, 1            ; swap as needed
                  mov al, [esi]
                  mov ah, [ecx]
                  mov [ecx], al
                  mov [esi], ah
                  add esi, 1
                .until (esi >= ecx)
            .endif
        .elseif dh==2  ; flag to save a byte
            stosb
            mov byte ptr [edi],0
        .elseif dh==4 || dh==5 ; hex, dl=number of bytes
            .if dh==4
                mov dh,7   ; upper case
            .else
                mov dh,7+32   ; lower case
            .endif
            shl dl,1    ; two hex characters per byte
            mov cl,dl
            shl cl,2    ; * 4 bits per hex character
            ror eax,cl  ; move to upper end of word
            .repeat
                rol eax,4   ; get a nibble
                mov cl,al
                and ecx,0fh
                .if cl>9
                    add cl,dh    ; get up to an 'A' or 'a'
                 .endif
                 add cl,48      ; convert to ascii
                 mov [edi],cl
                 inc edi
                 dec dl
            .until zero?
            mov [edi],dl    ; terminator
        .elseif dh==6       ; new lines
            mov ax,0a0dh
            .repeat
                stosw
                dec dl
            .until zero?
            mov byte ptr [edi],0
        .elseif dh==8
            .if sdword ptr eax > 0
                mov ecx,eax
                mov al,32
                rep stosb
            .endif
        .endif
        PrintxErrRet:
        pop fs:[0]   ; Uninstall per-thread error handler
        add esp,4
        .if printxeflg
            mov eax,offset printxerr1
            jmp printxerrorexit
        .endif
        mov printxptr,edi
        sub edi,printxstart
        mov printxsize,edi
    .else
        mov eax,offset printxerr2
        printxerrorexit:
        invoke MessageBox,0,eax,addr printxname,0
    .endif
	ret

PrintxExceptionHandler: ; PROC pExcept, pFrame, pContext, pDispatch
    mov eax,[esp+12]  ; pContext
    mov [eax].CONTEXT.regEip,offset PrintxErrRet
    mov eax,ExceptionContinueExecution
    mov printxeflg,1
    mov edi,printxptr       ; discard this string
    mov [edi],byte ptr 0
    ret

CheckPrintxAddress:
    pusha
    mov printxptr,eax
    mov printxstart,eax
    mov printxsize,0

    mov printxeflg,0 ;Install per-thread error handler from http://www.winasm.net/forum/index.php?showtopic=2082
    assume fs:nothing ;MASM assumes the use of this register to be ERROR by default
    push offset CheckPrintxAddressExceptionHandler
    push fs:[0]
    mov fs:[0],esp
    
    mov dword ptr [eax],0

    CheckPrintxAddressErrRet:
    pop fs:[0]   ; Uninstall per-thread error handler
    add esp,4
    .if printxeflg
        mov eax,offset printxerr3
        invoke MessageBox,0,eax,addr printxname,0
        mov printxptr,0     ; don't use this address!
        mov printxstart,0
    .endif
    popa
    cmp printxstart,0   ; return flag
    ret

CheckPrintxAddressExceptionHandler:
    mov eax,[esp+12]  ; pContext
    mov [eax].CONTEXT.regEip,offset CheckPrintxAddressErrRet
    mov eax,ExceptionContinueExecution
    mov printxeflg,1
    ret

printx macro BufferAddr,Things:Vararg ;; copy all Things to BufferAddr
  ;; copys text strings surrounded by double or single quotes to buffer BufferAddr
  ;; an entry of the characters dd means the next value is a dword to be converted to text
  ;; printxa (printx append) may be called to append more text to current buffer
  ;; printxf is the function version of printx, which returns BufferAddr
  ;; printxaf is the function version of printxa, append to buffer, returns BufferAddr
    push eax
    if @SizeStr(BufferAddr) gt 5
        ifidni @SubStr(BufferAddr,1,5),<addr >
            printxbuffaddr textequ @SubStr(BufferAddr,6)
            if (opattr(printxbuffaddr)) eq 98
                lea eax,printxbuffaddr
            else
                mov eax,@CatStr(<offset >,printxbuffaddr)
            endif
        elseif @SizeStr(BufferAddr) gt 7
            ifidni @SubStr(BufferAddr,1,7),<offset >
                printxbuffaddr textequ @SubStr(BufferAddr,8)
                if (opattr(printxbuffaddr)) eq 98
                    lea eax,printxbuffaddr
                else
                    mov eax,BufferAddr
                endif
            else
                opxttr = opattr(BufferAddr)
                if opxttr eq 0
                    %echo printx: error in output buffer specification - BufferAddr
                    .err    
                    exitm
                endif   
                mov eax,BufferAddr
            endif
        else
            mov eax,BufferAddr
        endif
    else
        mov eax,BufferAddr        
    endif
    call CheckPrintxAddress ; go see if ok, and save if so
    pop eax
    .if !zero?
        printxa Things  ;; go append to buffer
    .endif   ; else error with buffer address 
endm
printxf macro BufferAddr,Things:Vararg ;; printx function returns buffer address
    printx BufferAddr,Things
    exitm <printxstart>
endm
printxaf macro ThingsToPrint:VARARG ;; printx append function, returns buffer address
    printxa ThingsToPrint
    exitm <printxstart>
endm
printxa macro ThingsToPrint:VARARG  ;; printx appends text to buffer starting at printxptr
    local dxflg,argsize,first1,spec
    for arg,<ThingsToPrint>
        xxxxxxxxxxxxxxxxxxxxxxxx textequ <&arg> ;; for debug
        dxflg=0     ;; assume string
        argsize SIZESTR <&arg>
        if argsize eq 0
            goto DoneArg
        endif            
        pusha ;; save any registers that may be printed from later in this call
        first1 SUBSTR <&arg>,1,1
        if (@InStr(,</>,<%first1>))
            crcnt=0
            dxflg = 6
            forc slashx,<&arg> 
                ifidn <&slashx>,</>
                    crcnt = crcnt + 1
                endif
            endm
            mov dl,crcnt
            goto Doit
        endif
        if (@InStr(,<!"!'>,<%first1>))   ;; if single or double quote
            .data
               tmpoffset = $
               db arg,0
            .code
            mov esi,tmpoffset
            goto Doit
        endif
        if argsize gt 3
            spec SUBSTR <arg>,1,3
            ifidni spec,<dd >
                dxflg = 1   ;; signed decimal
            else        ;; have to break up due to bug in masm.  jwasm ok.
              ifIDNi spec,<du >
                dxflg = 3   ;; unsigned decimal
              else
                 ifidni spec,<hx >
                    ifidn @SubStr(arg,1,1),<H>
                        dxflg = 4   ;; upper case hex
                    else
                        dxflg = 5   ;; lower case hex
                    endif
                 else
                    ifidni spec,<xx >
                        dxflg = 8   ;; blanks
                    else
                        ifidni spec,<fl >
                            dxflg = 7   ;; floating point
                        endif
                    endif
                 endif
              endif
            endif
        endif
        if dxflg
            bcnt=4    ;; assume 4 bytes
            opxttr = opattr(@SubStr(arg,4))
            ;;opxx textequ %opxttr
            ;;%echo opp= opxx &arg
            if opxttr eq 110000y ;; 30h a register
                if @SizeStr(@SubStr(arg,4)) eq 2
                    if dxflg eq 1
                        movsx eax,@SubStr(arg,4)    ;; signed
                    else ;; good for both du and hx, types 3 and 4
                        movzx eax,@SubStr(arg,4)
                    endif
                    ifidni @SubStr(arg,5),<h>  ;; type doesn't work on byte registers in masm
                        bcnt=1   ;; a byte (ah,bh,ch,dh)
                    else
                        ifidni @SubStr(arg,5),<l>
                            bcnt=1  ;; a byte (al,bl,cl,dl)
                        else
                            bcnt=2    ; was a word (ax,bx,si, etc.)
                        endif
                    endif
                else
                    ;if dxflg eq 7
                    ;    fld @SubStr(arg,4)
                    ;else
                        mov eax,@SubStr(arg,4)
                    ;endif
                endif
            elseif opxttr eq 100100y ;; 24h immediate value
                mov eax,@SubStr(arg,4)
            elseif opxttr eq 101010y or opxttr eq 100010y or opxttr eq 98 ;; 22h or 2Ah a variable name or proc parameter
                bcnt = type @SubStr(arg,4)
                if bcnt eq 2
                    if dxflg eq 1
                        movsx eax,word ptr @SubStr(arg,4)
                    else
                        movzx eax,word ptr @SubStr(arg,4)
                    endif
                elseif bcnt eq 1
                    if dxflg eq 1
                        movsx eax,byte ptr @SubStr(arg,4)
                    else
                        movzx eax,byte ptr @SubStr(arg,4)
                    endif
                else
                    ;if dxflg eq 7
                    ;    fld @SubStr(arg,4)
                    ;else
                        mov eax,@SubStr(arg,4)
                        bcnt = 4
                    ;endif
                endif
            elseif opxttr eq 98  ;; e.g. a proc parameter
                if sizeof arg lt 4
                    movzx eax,arg
                else
                    mov eax,arg
                endif
            else
                mov eax,@SubStr(arg,4)  ;; will produce an invalid address error is not valid
            endif
            if (dxflg eq 4) or (dxflg eq 5)
                mov dl,bcnt
            endif
        else

            if argsize gt 3
                ifidni @SubStr(arg,1,3),<aa >
                    mov esi,@SubStr(arg,4)
                    goto Doit
                endif
            endif
            if argsize gt 5
                ifidni @SubStr(arg,1,5),<addr >
                    lea esi,@SubStr(arg,6)
                    goto Doit
                endif
            endif
yyyyyyyyyyyyyyyyyyyyyyyyyy = opattr(arg)                      
            if (opattr(arg)) eq 100100y   ;; immediate value
                if (arg gt 0) and (arg lt 256)
                    mov al,arg
                    dxflg = 2
                else
                    mov eax,arg     ;; just print a large value
                    dxflg = 1
                endif
                goto Doit      ;; default to signed
            elseif (opattr(arg)) eq 98  ;; e.g. a proc parameter
                ;if sizeof arg lt 4
                ;    movzx esi,arg
                ;else
                    mov esi,arg
                ;endif
                goto Doit
            endif
            ifidn @SubStr(arg,1,1),<[>
                mov esi,[arg]
            else
                lea esi,[arg]
            endif
        endif
      :Doit
        mov dh,dxflg    ;; type for printxproc
        call printxproc
        popa
     :DoneArg
    endm
endm

msgx macro args:vararg  ;; show results of printx in message box
    pusha
    ifnb <args>
        printx args
    endif
    ifdef ProgName
        mov edx,offset ProgName
    else
        xor edx,edx
    endif
    call DoMsgx
    popa
endm

.code
DoMsgx:  ;call with eax=address of message, edx=address of ProgName or zero.  msgx saves all regs
    mov eax,printxstart
    .if eax
        invoke MessageBox,0,eax,edx,0
    .endif
    ret
    
ifndef m2m
m2m MACRO M1:req, M2:req
	push M2
	pop  M1
ENDM
endif
